#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 13 10:07:01 2020

@author: siebert
"""
#%%
import matplotlib.pyplot as plt

plt.rcParams.update({'font.size': 14}) 
plt.rcParams.update(plt.rcParamsDefault)
 
import numpy as np
 
from IPython import get_ipython
 
import pandas as pd
from scipy import signal
from matplotlib import colors as mcolors
from scipy.signal import savgol_filter
from bisect import bisect_left, bisect_right

import configparser
from configparser import SafeConfigParser
import time 
from IPython import display
from pandas.tseries.offsets import DateOffset

from datetime import datetime, timedelta

import matplotlib.dates as mdates
from matplotlib.dates import DateFormatter

import os

from pathlib import Path

import colorama
from colorama import Fore
from colorama import Style

import warnings
warnings.filterwarnings("ignore")

colorama.init()

class bcolors:
    OK = '\033[92m' #GREEN
    WARNING = '\033[93m' #YELLOW
    FAIL = '\033[91m' #RED
    RESET = '\033[0m' #RESET COLOR

# Define values
Rv  = 461
RL  = 287
cp  = 1004                                  
T0  = 273.15 
p0  = 1013                                
L   = 2.5e6
g   = 9.81
rho_s = 1.2

def ES(T):    
    return 6.1 * np.exp(17.62 * T /(243.12 + T))  

def e(ES,rH):
    return rH / 100 * (ES * 100)

def absH(rH,ES,T):
    return (rH * ES) / Rv / (T+T0)

def mixRatio(rH,ES,p):
    return rH/100 * ES / p * RL/Rv * 1e-3

def potT(T,p):
    return (T + T0) * (p0/p)**(RL / cp)

def zb(p,p_GND,T_GND,RL,g):
    return RL * (T_GND + 273.15) / g * np.log(p_GND / p)

 
get_ipython().run_line_magic('matplotlib', 'qt5')

#%%            Read configuration ....

config   = "Config_MOSAiC_public.ini"
 
## MOSAiC 
# Select flights 

# DEFAULT  = '20200629-1'
# DEFAULT  = '20200629-2'
# DEFAULT  = '20200719'
# DEFAULT  = '20200714'

# DEFAULT  = '20200710'
# DEFAULT  = '20200715'
# DEFAULT = '20200719'
# DEFAULT  = '20200720-1'
# DEFAULT  = '20200720-2'
# DEFAULT  = '20200721'
# DEFAULT  = '20200721-2'
# DEFAULT  = '20200722'

# DEFAULT  = '20200723'
# DEFAULT  = '20200724'
# DEFAULT  = '20200725'
# DEFAULT  = '20200726'
DEFAULT  = '20200727'

parser = SafeConfigParser()
parser.read(config)

FILE_XDAT  =  os.path.join(parser.get(DEFAULT, 'PATH') , parser.get(DEFAULT, 'FILE_XDAT' ))
DATE       =  parser.get(DEFAULT, 'DATE')

print("###############################################################################################")
print(Fore.RED + Style.BRIGHT +"Date       = " , DATE )
print("XDAT File  = " , FILE_XDAT + Style.RESET_ALL)
print("##############################################################################################")

year_off   = DATE[0:4]
month_off  = DATE[5:7]
day_off    = DATE[8:10]
date_off   = year_off + " - " + month_off + " - " + day_off

#%%      #########  Read and process *.xdat file    

my_cols     = [str(i) for i in range(32)] 

col_xdat    = ["FN","frametim","pres","temp","rh",
              "SOD","ALT","ID","IN","SEQU","HEAD","VELO","PITCH",
              "ROLL","TEMP","ALPHA","VISIB","VAL","CHK"]  

# defines the first valid data, default = 0

RM_N = np.int(parser.get(DEFAULT, 'N'))
print( 'start reading at line ', RM_N)
 
df         = pd.read_csv(FILE_XDAT,sep='[\t\x20]\x20*' ,skiprows=RM_N,header=0,names = col_xdat, index_col= False , engine = "python")   
 
a = len(df)
print("Read ", a ," lines")
 
df         = df.fillna(method='ffill') # invalid numbers will be replace by the previous one

print( a- len(df), Fore.RED + Style.BRIGHT + " bad lines with NaN removed" + Style.RESET_ALL)

# convert all data to numeric
df         = df.apply(pd.to_numeric, errors='coerce', axis=1)
 
#%%
# frequently, the GPS sequence was not valid from the first launch, therefore the time information is missing (second of day, SOD)
# find the first valid SOD and replace all SOD == 0

bb = np.where(df.SOD != 0)[0]

if  bb.size  == 0:
    print("No valid GPS == >> No SOD !! ")
    idx_min = 0
    df.SOD[0] = parser.get(DEFAULT, 'SOD_0')
    print("Got", parser.get(DEFAULT, 'SOD_0'), "from ini file")
else:
    print("at least one valid SOD value found ...")
    idx_min   = bb.min()
    df.SOD[0] = df.SOD[idx_min] - idx_min
    print("1st valid SOD point = ",df.SOD[idx_min], "at Index", idx_min)


if idx_min < 10:   
    idx_min = 10

for j in range(idx_min):
    df.SOD[df.SOD == j] = df.SOD.shift(periods = 1) + 1 

# calculate the first SOD to set the correct starting time 
df        = df.set_index(pd.to_datetime(df.SOD, unit='s', errors='coerce'))
 
df.index  = df.index + DateOffset(year = int(year_off),months=int(month_off) -1 , day = int(day_off))
   
# index is now the correct date and time of the record, checked with transponder via flightradar24


# select sub-record as defined in the ini-file for further processing

print("Start record :", df.index[0])
print("End record   :", df.index.max())

print("size of complete dataframe ",len(df))

start = parser.get(DEFAULT, 'start')
end   = parser.get(DEFAULT, 'end')
df    = df[ start: end]

print("size of selected subdataframe ",len(df))
print("")
print(Fore.RED + Style.BRIGHT + "  Subsection from   "+ Style.RESET_ALL , start, " to " , end )

a = len(df)

print("")

# check if index is pure monotonic increasing

if df.index.is_unique == False:
    print(Fore.RED + Style.BRIGHT +"duplicate indices found ==>> will be removed")
    df = df[~df.index.duplicated(keep='first')]
    b = len(df)
    if df.index.is_unique == True:
        print(a-b,"duplicate indices removed" +Style.RESET_ALL )
else: print("no duplicate indices found")

 

#%% 
dT               = df.temp.diff().abs()     # detect specific spikes with amplitude of increments > 10K (artificial)
df.temp[dT >= 10] = df.temp - dT
#%% 

              
p_GND = df.pres[0:60].max() # define maximum static pressure of first minute after "switch on" as ground value

# define height period for estimating "ground values" of temperature and humidity after the sensor has been launched 

p_low = 0.2   # lower pressure boundary (about 0.8 m above ground level)
p_up  = 0.5   # upper pressure boundary (about 4.0 m above ground level)

idx_p = np.where( (df.pres <= (p_GND - p_low)) & (df.pres >= (p_GND - p_up)) ) # indices of all data between 0.5 and 4.0 m above ground level

T_GND     = df.temp[idx_p[0]].mean()
rH_GND    = df.rh  [idx_p[0]].mean()

## Read ground values from MetTower (MOSAiC) as stored in the ini-file

T_TWR_GND  = np.float16(parser.get(DEFAULT, 'T_Met'))
rH_TWR_GND = np.float16(parser.get(DEFAULT, 'rH_Met'))
p_TWR_GND  = np.float32(parser.get(DEFAULT, 'p_Met'))

print('')
print("p_GND  = " , p_GND)
print(Fore.RED + Style.BRIGHT + "Ground Values from Std Met Probe for height range 1.6 - 4.0 m"+ Style.RESET_ALL )
print("#############################################################")
print("T_GND  = " , T_GND)
print("rH_GND = " , rH_GND)
print('')
print(Fore.RED + Style.BRIGHT + "Ground Values from Met Tower"+ Style.RESET_ALL )
print()
print("#############################################################")
print("p_TWR_GND  = " , p_TWR_GND)
print("T_TWR_GND  = " , T_TWR_GND)
print("rH_TWR_GND = " , rH_TWR_GND)
print('')

          
df.rh[df.rh > 100] = 100 # set all humidity readings > 100% to exactly 100%
df = df.assign(rh_c = df.rh)

df.pres = df.pres - (p_GND - p_TWR_GND) # correct pressure offset assuming the tower value is correct
print('')
print(Fore.RED + Style.BRIGHT +"pressure correction done" + Style.RESET_ALL)
     
df = df.assign(temp_c = df.temp - (T_GND - T_TWR_GND)) # correct temperature offset assuming the tower value is correct
print(Fore.RED + Style.BRIGHT +"temperature correction done" + Style.RESET_ALL)

df        = df.assign(zb   = zb(df.pres,df.pres.max(),(T_TWR_GND),RL,g))       # define barometric height according ICAO
df        = df.assign(rho  = (df.pres * 100) / (RL *( df.temp_c + 273.15) ) )  # define air density

#%%
#####   recover differential pressure
# Background: 
# - a polynomial "correction" has been applied to the raw values of the differential pressure device which then have been transmitted to the ground station
# - only the final velocity data have been stored but after a re-calibration, the polynomial fit was considered as faulty
# - the raw values can be restored by applying a "look-up table"  
# - reliable velocity data has been estimated by a comparison with a new miniature ultrasonic anemometer (miniSonic)
    
df.VELO[df.VELO > 20] = 0  # values above 20 m/s are considered as faulty
df.ROLL               = df.ROLL * (-1)
lookup_file = "lookup_table.txt" 
col_lookup  = ["VEL","BIN","HEX"]
df_lookup   = pd.read_csv(lookup_file,sep='[\t\x20]\x20*' ,skiprows=1,header=(0),names = col_lookup, engine = "python")  

print(Fore.RED + Style.BRIGHT +"read lookup table" + Style.RESET_ALL)

df    = df.assign(dp   = df.VELO * 0)
df    = df.assign(vdp  = df.VELO * 0)
df    = df.assign(vc   = df.VELO * 0)

df    = df.assign(pBin = df.VELO * 0)
df    = df.assign(VCal = df.VELO * 0)
df    = df.assign(flag = df.VELO * 0)
  
for i in range(len(df)):

    pr         = df_lookup.iloc[(df_lookup['VEL'] - df.VELO[i]).abs().argsort()[:1]]
    df.pBin[i] = int(pr.BIN)                                 # raw binary values from the pressure device    
    df.dp[i]   = 2.5 * (517/26214. * (pr.BIN - 3277)) - 0.44 # new differential pressure dp according calibration with miniSonic
    
dp_limit = np.float16(parser.get(DEFAULT, 'dp_limit'))

print(Fore.RED + Style.BRIGHT + "  recovering of dp done ...  "+ Style.RESET_ALL )
    
#%%

df.dp  = df.dp[(df.dp > 0)  & (df.dp < dp_limit)]

 
df.dp  = df.dp.fillna(0)
df.dp  = df.dp - 7 # typical offset in container with zero flow 
df.vdp = np.sqrt(( 2 * df.dp / df.rho)) 
    
df.vdp = df.vdp.fillna(0)

status = (int(parser.get(DEFAULT, 'status')) ) # see ini-file, status == 0 means no trustworthy wind data

if   status == 0:
    print(Fore.RED + Style.BRIGHT + 'no valid wind data'+ Style.RESET_ALL)
    df.vdp = np.NAN
else:
    df.vdp  = df.vdp

df.HEAD[df.HEAD > 360] = np.NaN
df.HEAD                = df.HEAD.interpolate(method='linear', axis=0).bfill().rolling(5).median()

#%% writing csv data

col_list =["pres", "temp_c", "rh_c", "vdp", "HEAD", "zb"]
header = [ "p", "T", "rH", "U", "D","zb"]

base = os.path.splitext(FILE_XDAT)[0]
FILE_OUT = (base + '.csv2')

df[col_list].to_csv(FILE_OUT,index = True, float_format='%.2f', header = header)

# finished ....


 